/*
 * JIntellitype
 * -----------------
 * Copyright 2005-2008 Emil A. Lefkof III, Melloware Inc.
 *
 * I always give it my best shot to make a program useful and solid, but
 * remeber that there is absolutely no warranty for using this program as
 * stated in the following terms:
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.melloware.jintellitype;

import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

/**
 * JIntellitype A Java Implementation for using the Windows API Intellitype
 * commands and the RegisterHotKey and UnRegisterHotkey API calls for globally
 * responding to key events. Intellitype are commands that are using for Play,
 * Stop, Next on Media keyboards or some laptops that have those special keys.
 * <p>
 * JIntellitype class that is used to call Windows API calls using the
 * JIntellitype.dll.
 * <p>
 * This file comes with native code in JINTELLITYPE.DLL The DLL should go in
 * C:/WINDOWS/SYSTEM or in your current directory
 * <p>
 * <p>
 * Copyright (c) 1999-2008 Melloware, Inc. <http://www.melloware.com>
 * @author Emil A. Lefkof III <info@melloware.com>
 * @version 1.3.1
 */
public final class JIntellitype {

	/**
	 * Static variable to hold singleton.
	 */
	private static JIntellitype jintellitype = null;

	/**
	 * Static variable to hold the libary location if set
	 */
	private static String libraryLocation = null;

	/**
	 * Listeners collection for Hotkey events
	 * <p>
	 * HotkeyListener
	 */
	private final List hotkeyListeners = new ArrayList();

	/**
	 * Listeners collection for Hotkey events
	 * <p>
	 * IntellitypeListener
	 */
	private final List intellitypeListeners = new ArrayList();

	/**
	 * Handler is used by JNI code to keep different JVM instances separate
	 */
	private final long handler = 0;

	/**
	 * Map containing key->keycode mapping
	 * <p>
	 * String, Integer
	 * @see #registerHotKey(int, String)
	 * @see #getKey2KeycodeMapping()
	 */
	private final HashMap keycodeMap;

	/**
	 * Private Constructor to prevent instantiation. Initialize the library for
	 * calling.
	 */
	private JIntellitype() {
		try {
			// Load JNI library
			System.loadLibrary("JIntellitype");
		} catch (Throwable ex) {
			try {
				if (getLibraryLocation() != null) {
					System.load(getLibraryLocation());
				} else {
					String jarPath = "com/melloware/jintellitype/";
					String tmpDir = System.getProperty("java.io.tmpdir");
					try {
						String dll = "JIntellitype.dll";
						fromJarToFs(jarPath + dll, tmpDir + dll);
						System.load(tmpDir + dll);
					} catch (UnsatisfiedLinkError e) {
						String dll = "JIntellitype64.dll";
						fromJarToFs(jarPath + dll, tmpDir + dll);
						System.load(tmpDir + dll);
					}
				}
			} catch (Throwable ex2) {
				throw new RuntimeException(
						"Could not load JIntellitype.dll from local file system or from inside JAR",
						ex2);
			}
		}

		initializeLibrary();
		this.keycodeMap = getKey2KeycodeMapping();
	}

	/**
	 * Pulls a file out of the JAR and puts it on the File Path.
	 * <p>
	 * @param jarPath the path to the JAR
	 * @param filePath the file path to extract to
	 * @throws IOException if any IO error occurs
	 */
	private void fromJarToFs(String jarPath, String filePath)
			throws IOException {
		File file = new File(filePath);
		if (file.exists()) {
			boolean success = file.delete();
			if (!success) {
				throw new IOException("couldn't delete " + filePath);
			}
		}
		InputStream is = null;
		OutputStream os = null;
		try {
			is = ClassLoader.getSystemClassLoader()
					.getResourceAsStream(jarPath);
			os = new FileOutputStream(filePath);
			byte[] buffer = new byte[8192];
			int bytesRead;
			while ((bytesRead = is.read(buffer)) != -1) {
				os.write(buffer, 0, bytesRead);
			}
		} finally {
			if (is != null) {
				is.close();
			}
			if (os != null) {
				os.close();
			}
		}
	}

	/**
	 * Gets the singleton instance of the JIntellitype object.
	 * <p>
	 * But the possibility of creation of more instance is only before the
	 * instance is created. Since all code defined inside getInstance method is
	 * in the synchronized block, even the subsequent requests will also come and
	 * wait in the synchronized block. This is a performance issue. The same can
	 * be solved using double-checked lock. Following is the implementation of
	 * Singleton with lazy initialization and double-checked lock.
	 * <p>
	 * @return an instance of JIntellitype class
	 */
	public static JIntellitype getInstance() {
		if (jintellitype == null) {
			jintellitype = new JIntellitype();
		}
		return jintellitype;
	}

	/**
	 * Adds a listener for hotkeys.
	 * <p>
	 * @param listener the HotKeyListener to be added
	 */
	public synchronized void addHotKeyListener(HotkeyListener listener) {
		hotkeyListeners.add(listener);
	}

	/**
	 * Adds a listener for intellitype commands.
	 * <p>
	 * @param listener the IntellitypeListener to be added
	 */
	public synchronized void addIntellitypeListener(IntellitypeListener listener) {
		intellitypeListeners.add(listener);
	}

	/**
	 * Cleans up all resources used by JIntellitype.
	 */
	public void cleanUp() {
		terminate();
	}

	/**
	 * Registers a Hotkey with windows. This combination will be responded to by
	 * all registered HotKeyListeners. Uses the JIntellitypeConstants for MOD,
	 * ALT, CTRL, and WINDOWS keys.
	 * <p>
	 * @param identifier a unique identifier for this key combination
	 * @param modifier MOD_SHIFT, MOD_ALT, MOD_CONTROL, MOD_WIN from
	 *           JIntellitypeConstants, or 0 if no modifier needed
	 * @param keycode the key to respond to in Ascii integer, 65 for A
	 */
	public void registerHotKey(int identifier, int modifier, int keycode) {
		regHotKey(identifier, modifier, keycode);
	}

	/**
	 * Registers a Hotkey with windows. This combination will be responded to by
	 * all registered HotKeyListeners. Use the Swing InputEvent constants from
	 * java.awt.InputEvent.
	 * <p>
	 * @param identifier a unique identifier for this key combination
	 * @param modifier InputEvent.SHIFT_MASK, InputEvent.ALT_MASK,
	 *           InputEvent.CTRL_MASK, or 0 if no modifier needed
	 * @param keycode the key to respond to in Ascii integer, 65 for A
	 */
	public void registerSwingHotKey(int identifier, int modifier, int keycode) {
		regHotKey(identifier, swingToIntelliType(modifier), keycode);
	}

	/**
	 * Registers a Hotkey with windows. This combination will be responded to by
	 * all registered HotKeyListeners. Use the identifiers CTRL, SHIFT, ALT
	 * and/or WIN.
	 * <p>
	 * @param identifier a unique identifier for this key combination
	 * @param modifierAndKeyCode String with modifiers separated by + and keycode
	 *           (e.g. CTRL+SHIFT+A)
	 * @see #registerHotKey(int, int, int)
	 * @see #registerSwingHotKey(int, int, int)
	 */
	public void registerHotKey(int identifier, String modifierAndKeyCode) {
		String[] split = modifierAndKeyCode.split("\\+"); //$NON-NLS-1$
		if (split.length < 2) {
			throw new IllegalArgumentException("delimiter \"+\" not found."); //$NON-NLS-1$
		}
		int mask = 0;
		for (int i = 0; i < split.length - 1; i++) {
			if ("ALT".equalsIgnoreCase(split[i])) { //$NON-NLS-1$
				mask += JIntellitypeConstants.MOD_ALT;
			} else if ("CTRL".equalsIgnoreCase(split[i]) //$NON-NLS-1$
					|| "CONTROL".equalsIgnoreCase(split[i])) { //$NON-NLS-1$
				mask += JIntellitypeConstants.MOD_CONTROL;
			} else if ("SHIFT".equalsIgnoreCase(split[i])) { //$NON-NLS-1$
				mask += JIntellitypeConstants.MOD_SHIFT;
			} else if ("WIN".equalsIgnoreCase(split[i])) { //$NON-NLS-1$
				mask += JIntellitypeConstants.MOD_WIN;
			} else {
				throw new IllegalArgumentException("Invalid modifier:" //$NON-NLS-1$
						+ split[i]);
			}
		}
		String key = split[split.length - 1];
		if (key.length() == 1) {
			registerHotKey(identifier, mask, key.charAt(0));

		} else if (key.startsWith("0x") && key.length() > 3) { //$NON-NLS-1$
			try {
				int code = Integer.parseInt(key.substring(2), 16);
				registerHotKey(identifier, mask, code);
			} catch (NumberFormatException e) {
				throw new IllegalArgumentException(e.toString());
			}
		} else if (keycodeMap.containsKey(key.toLowerCase())) {
			Integer code = (Integer) keycodeMap.get(key.toLowerCase());
			registerHotKey(identifier, mask, code.intValue());
		} else {
			throw new IllegalArgumentException(
					"Invalid key combination format=" + key); //$NON-NLS-1$
		}
	}

	/**
	 * Removes a listener for hotkeys.
	 */
	public synchronized void removeHotKeyListener(HotkeyListener listener) {
		hotkeyListeners.remove(listener);
	}

	/**
	 * Removes a listener for intellitype commands.
	 */
	public synchronized void removeIntellitypeListener(
			IntellitypeListener listener) {
		intellitypeListeners.remove(listener);
	}

	/**
	 * Unregisters a previously registered Hotkey identified by its unique
	 * identifier.
	 * <p>
	 * @param identifier the unique identifer of this Hotkey
	 */
	public void unregisterHotKey(int identifier) {
		unregHotKey(identifier);
	}

	/**
	 * Checks to see if this application is already running.
	 * <p>
	 * @param appTitle the name of the application to check for
	 * @return true if running, false if not running
	 */
	public static boolean checkInstanceAlreadyRunning(String appTitle) {
		return getInstance().isRunning(appTitle);
	}

	/**
	 * Checks to make sure the OS is a Windows flavor and that the JIntellitype
	 * DLL is found in the path and the JDK is 32 bit not 64 bit. The DLL
	 * currently only supports 32 bit JDK.
	 * <p>
	 * @return true if Jintellitype may be used, false if not
	 */
	public static boolean isJIntellitypeSupported() {
		boolean result = false;
		String os = "none";

		try {
			os = System.getProperty("os.name").toLowerCase();
		} catch (SecurityException ex) {
			// we are not allowed to look at this property
			System.err
					.println("Caught a SecurityException reading the system property "
							+ "'os.name'; the SystemUtils property value will default to null.");
		}

		// only works on Windows OS currently
		if (os.startsWith("windows")) {
			// try an get the instance and if it succeeds then return true
			try {
				getInstance();
				result = true;
			} catch (Exception e) {
				result = false;
			}
		}

		return result;
	}

	/**
	 * Gets the libraryLocation.
	 * <p>
	 * @return Returns the libraryLocation.
	 */
	public static String getLibraryLocation() {
		return libraryLocation;
	}

	/**
	 * Sets the libraryLocation.
	 * <p>
	 * @param libraryLocation The libraryLocation to set.
	 */
	public static void setLibraryLocation(String libraryLocation) {
		final File dll = new File(libraryLocation);
		if (!dll.isAbsolute()) {
			JIntellitype.libraryLocation = dll.getAbsolutePath();
		} else {
			// absolute path, no further calculation needed
			JIntellitype.libraryLocation = libraryLocation;
		}
	}

	/**
	 * Notifies all listeners that Hotkey was pressed.
	 * <p>
	 * @param identifier the unique identifier received
	 */
	protected synchronized void onHotKey(final int identifier) {
		for (Iterator itr = hotkeyListeners.iterator(); itr.hasNext();) {
			HotkeyListener listener = (HotkeyListener) itr.next();
			listener.onHotKey(identifier);
		}
	}

	/**
	 * Notifies all listeners that Intellitype command was received.
	 * <p>
	 * @param command the unique WM_APPCOMMAND received
	 */
	protected synchronized void onIntellitype(final int command) {
		for (Iterator itr = intellitypeListeners.iterator(); itr.hasNext();) {
			IntellitypeListener intellitypeListener = (IntellitypeListener) itr
					.next();
			intellitypeListener.onIntellitype(command);
		}
	}

	/**
	 * Swing modifier value to Jintellipad conversion. If no conversion needed
	 * just return the original value. This lets users pass either the original
	 * JIntellitype constants or Swing InputEvent constants.
	 * <p>
	 * @param swingKeystrokeModifier the Swing KeystrokeModifier to check
	 * @return Jintellitype the JIntellitype modifier value
	 */
	protected static int swingToIntelliType(int swingKeystrokeModifier) {
		int mask = 0;
		if ((swingKeystrokeModifier & InputEvent.SHIFT_MASK) == InputEvent.SHIFT_MASK) {
			mask += JIntellitypeConstants.MOD_SHIFT;
		}
		if ((swingKeystrokeModifier & InputEvent.ALT_MASK) == InputEvent.ALT_MASK) {
			mask += JIntellitypeConstants.MOD_ALT;
		}
		if ((swingKeystrokeModifier & InputEvent.CTRL_MASK) == InputEvent.CTRL_MASK) {
			mask += JIntellitypeConstants.MOD_CONTROL;
		}
		if ((swingKeystrokeModifier & InputEvent.SHIFT_DOWN_MASK) == InputEvent.SHIFT_DOWN_MASK) {
			mask += JIntellitypeConstants.MOD_SHIFT;
		}
		if ((swingKeystrokeModifier & InputEvent.ALT_DOWN_MASK) == InputEvent.ALT_DOWN_MASK) {
			mask += JIntellitypeConstants.MOD_ALT;
		}
		if ((swingKeystrokeModifier & InputEvent.CTRL_DOWN_MASK) == InputEvent.CTRL_DOWN_MASK) {
			mask += JIntellitypeConstants.MOD_CONTROL;
		}
		return mask;
	}

	/**
	 * Puts all constants from {@link java.awt.event.KeyEvent} in a keycodeMap.
	 * The key is the lower case form of it.
	 * @return Map containing key->keycode mapping DOCU Now enables the user to
	 *         use all keys specified here instead of just [A-Z],[0-9] as before
	 */
	private HashMap getKey2KeycodeMapping() {
		HashMap map = new HashMap();

		put(map, "first", KeyEvent.KEY_FIRST);
		put(map, "last", KeyEvent.KEY_LAST);
		put(map, "typed", KeyEvent.KEY_TYPED);
		put(map, "pressed", KeyEvent.KEY_PRESSED);
		put(map, "released", KeyEvent.KEY_RELEASED);
		put(map, "enter", 13);
		put(map, "back_space", KeyEvent.VK_BACK_SPACE);
		put(map, "tab", KeyEvent.VK_TAB);
		put(map, "cancel", KeyEvent.VK_CANCEL);
		put(map, "clear", KeyEvent.VK_CLEAR);
		put(map, "pause", KeyEvent.VK_PAUSE);
		put(map, "caps_lock", KeyEvent.VK_CAPS_LOCK);
		put(map, "escape", KeyEvent.VK_ESCAPE);
		put(map, "space", KeyEvent.VK_SPACE);
		put(map, "page_up", KeyEvent.VK_PAGE_UP);
		put(map, "page_down", KeyEvent.VK_PAGE_DOWN);
		put(map, "end", KeyEvent.VK_END);
		put(map, "home", KeyEvent.VK_HOME);
		put(map, "left", KeyEvent.VK_LEFT);
		put(map, "up", KeyEvent.VK_UP);
		put(map, "right", KeyEvent.VK_RIGHT);
		put(map, "down", KeyEvent.VK_DOWN);
		put(map, "comma", 188);
		put(map, "minus", 109);
		put(map, "period", 110);
		put(map, "slash", 191);
		put(map, "0", KeyEvent.VK_0);
		put(map, "1", KeyEvent.VK_1);
		put(map, "2", KeyEvent.VK_2);
		put(map, "3", KeyEvent.VK_3);
		put(map, "4", KeyEvent.VK_4);
		put(map, "5", KeyEvent.VK_5);
		put(map, "6", KeyEvent.VK_6);
		put(map, "7", KeyEvent.VK_7);
		put(map, "8", KeyEvent.VK_8);
		put(map, "9", KeyEvent.VK_9);
		put(map, "semicolon", 186);
		put(map, "equals", 187);
		put(map, "a", KeyEvent.VK_A);
		put(map, "b", KeyEvent.VK_B);
		put(map, "c", KeyEvent.VK_C);
		put(map, "d", KeyEvent.VK_D);
		put(map, "e", KeyEvent.VK_E);
		put(map, "f", KeyEvent.VK_F);
		put(map, "g", KeyEvent.VK_G);
		put(map, "h", KeyEvent.VK_H);
		put(map, "i", KeyEvent.VK_I);
		put(map, "j", KeyEvent.VK_J);
		put(map, "k", KeyEvent.VK_K);
		put(map, "l", KeyEvent.VK_L);
		put(map, "m", KeyEvent.VK_M);
		put(map, "n", KeyEvent.VK_N);
		put(map, "o", KeyEvent.VK_O);
		put(map, "p", KeyEvent.VK_P);
		put(map, "q", KeyEvent.VK_Q);
		put(map, "r", KeyEvent.VK_R);
		put(map, "s", KeyEvent.VK_S);
		put(map, "t", KeyEvent.VK_T);
		put(map, "u", KeyEvent.VK_U);
		put(map, "v", KeyEvent.VK_V);
		put(map, "w", KeyEvent.VK_W);
		put(map, "x", KeyEvent.VK_X);
		put(map, "y", KeyEvent.VK_Y);
		put(map, "z", KeyEvent.VK_Z);
		put(map, "open_bracket", 219);
		put(map, "back_slash", 220);
		put(map, "close_bracket", 221);
		put(map, "numpad0", KeyEvent.VK_NUMPAD0);
		put(map, "numpad1", KeyEvent.VK_NUMPAD1);
		put(map, "numpad2", KeyEvent.VK_NUMPAD2);
		put(map, "numpad3", KeyEvent.VK_NUMPAD3);
		put(map, "numpad4", KeyEvent.VK_NUMPAD4);
		put(map, "numpad5", KeyEvent.VK_NUMPAD5);
		put(map, "numpad6", KeyEvent.VK_NUMPAD6);
		put(map, "numpad7", KeyEvent.VK_NUMPAD7);
		put(map, "numpad8", KeyEvent.VK_NUMPAD8);
		put(map, "numpad9", KeyEvent.VK_NUMPAD9);
		put(map, "multiply", KeyEvent.VK_MULTIPLY);
		put(map, "add", KeyEvent.VK_ADD);
		put(map, "separator", KeyEvent.VK_SEPARATOR);
		put(map, "subtract", KeyEvent.VK_SUBTRACT);
		put(map, "decimal", KeyEvent.VK_DECIMAL);
		put(map, "divide", KeyEvent.VK_DIVIDE);
		put(map, "delete", 46);
		put(map, "num_lock", KeyEvent.VK_NUM_LOCK);
		put(map, "scroll_lock", KeyEvent.VK_SCROLL_LOCK);
		put(map, "f1", KeyEvent.VK_F1);
		put(map, "f2", KeyEvent.VK_F2);
		put(map, "f3", KeyEvent.VK_F3);
		put(map, "f4", KeyEvent.VK_F4);
		put(map, "f5", KeyEvent.VK_F5);
		put(map, "f6", KeyEvent.VK_F6);
		put(map, "f7", KeyEvent.VK_F7);
		put(map, "f8", KeyEvent.VK_F8);
		put(map, "f9", KeyEvent.VK_F9);
		put(map, "f10", KeyEvent.VK_F10);
		put(map, "f11", KeyEvent.VK_F11);
		put(map, "f12", KeyEvent.VK_F12);
		put(map, "f13", KeyEvent.VK_F13);
		put(map, "f14", KeyEvent.VK_F14);
		put(map, "f15", KeyEvent.VK_F15);
		put(map, "f16", KeyEvent.VK_F16);
		put(map, "f17", KeyEvent.VK_F17);
		put(map, "f18", KeyEvent.VK_F18);
		put(map, "f19", KeyEvent.VK_F19);
		put(map, "f20", KeyEvent.VK_F20);
		put(map, "f21", KeyEvent.VK_F21);
		put(map, "f22", KeyEvent.VK_F22);
		put(map, "f23", KeyEvent.VK_F23);
		put(map, "f24", KeyEvent.VK_F24);
		put(map, "printscreen", 44);
		put(map, "insert", 45);
		put(map, "help", 47);
		put(map, "meta", KeyEvent.VK_META);
		put(map, "back_quote", KeyEvent.VK_BACK_QUOTE);
		put(map, "quote", KeyEvent.VK_QUOTE);
		put(map, "kp_up", KeyEvent.VK_KP_UP);
		put(map, "kp_down", KeyEvent.VK_KP_DOWN);
		put(map, "kp_left", KeyEvent.VK_KP_LEFT);
		put(map, "kp_right", KeyEvent.VK_KP_RIGHT);
		put(map, "dead_grave", KeyEvent.VK_DEAD_GRAVE);
		put(map, "dead_acute", KeyEvent.VK_DEAD_ACUTE);
		put(map, "dead_circumflex", KeyEvent.VK_DEAD_CIRCUMFLEX);
		put(map, "dead_tilde", KeyEvent.VK_DEAD_TILDE);
		put(map, "dead_macron", KeyEvent.VK_DEAD_MACRON);
		put(map, "dead_breve", KeyEvent.VK_DEAD_BREVE);
		put(map, "dead_abovedot", KeyEvent.VK_DEAD_ABOVEDOT);
		put(map, "dead_diaeresis", KeyEvent.VK_DEAD_DIAERESIS);
		put(map, "dead_abovering", KeyEvent.VK_DEAD_ABOVERING);
		put(map, "dead_doubleacute", KeyEvent.VK_DEAD_DOUBLEACUTE);
		put(map, "dead_caron", KeyEvent.VK_DEAD_CARON);
		put(map, "dead_cedilla", KeyEvent.VK_DEAD_CEDILLA);
		put(map, "dead_ogonek", KeyEvent.VK_DEAD_OGONEK);
		put(map, "dead_iota", KeyEvent.VK_DEAD_IOTA);
		put(map, "dead_voiced_sound", KeyEvent.VK_DEAD_VOICED_SOUND);
		put(map, "dead_semivoiced_sound", KeyEvent.VK_DEAD_SEMIVOICED_SOUND);
		put(map, "ampersand", KeyEvent.VK_AMPERSAND);
		put(map, "asterisk", KeyEvent.VK_ASTERISK);
		put(map, "quotedbl", KeyEvent.VK_QUOTEDBL);
		put(map, "less", KeyEvent.VK_LESS);
		put(map, "greater", KeyEvent.VK_GREATER);
		put(map, "braceleft", KeyEvent.VK_BRACELEFT);
		put(map, "braceright", KeyEvent.VK_BRACERIGHT);
		put(map, "at", KeyEvent.VK_AT);
		put(map, "colon", KeyEvent.VK_COLON);
		put(map, "circumflex", KeyEvent.VK_CIRCUMFLEX);
		put(map, "dollar", KeyEvent.VK_DOLLAR);
		put(map, "euro_sign", KeyEvent.VK_EURO_SIGN);
		put(map, "exclamation_mark", KeyEvent.VK_EXCLAMATION_MARK);
		put(map, "inverted_exclamation_mark",
				KeyEvent.VK_INVERTED_EXCLAMATION_MARK);
		put(map, "left_parenthesis", KeyEvent.VK_LEFT_PARENTHESIS);
		put(map, "number_sign", KeyEvent.VK_NUMBER_SIGN);
		put(map, "plus", KeyEvent.VK_PLUS);
		put(map, "right_parenthesis", KeyEvent.VK_RIGHT_PARENTHESIS);
		put(map, "underscore", KeyEvent.VK_UNDERSCORE);
		put(map, "context_menu", KeyEvent.VK_CONTEXT_MENU);
		put(map, "final", KeyEvent.VK_FINAL);
		put(map, "convert", KeyEvent.VK_CONVERT);
		put(map, "nonconvert", KeyEvent.VK_NONCONVERT);
		put(map, "accept", KeyEvent.VK_ACCEPT);
		put(map, "modechange", KeyEvent.VK_MODECHANGE);
		put(map, "kana", KeyEvent.VK_KANA);
		put(map, "kanji", KeyEvent.VK_KANJI);
		put(map, "alphanumeric", KeyEvent.VK_ALPHANUMERIC);
		put(map, "katakana", KeyEvent.VK_KATAKANA);
		put(map, "hiragana", KeyEvent.VK_HIRAGANA);
		put(map, "full_width", KeyEvent.VK_FULL_WIDTH);
		put(map, "half_width", KeyEvent.VK_HALF_WIDTH);
		put(map, "roman_characters", KeyEvent.VK_ROMAN_CHARACTERS);
		put(map, "all_candidates", KeyEvent.VK_ALL_CANDIDATES);
		put(map, "previous_candidate", KeyEvent.VK_PREVIOUS_CANDIDATE);
		put(map, "code_input", KeyEvent.VK_CODE_INPUT);
		put(map, "japanese_katakana", KeyEvent.VK_JAPANESE_KATAKANA);
		put(map, "japanese_hiragana", KeyEvent.VK_JAPANESE_HIRAGANA);
		put(map, "japanese_roman", KeyEvent.VK_JAPANESE_ROMAN);
		put(map, "kana_lock", KeyEvent.VK_KANA_LOCK);
		put(map, "input_method_on_off", KeyEvent.VK_INPUT_METHOD_ON_OFF);
		put(map, "cut", KeyEvent.VK_CUT);
		put(map, "copy", KeyEvent.VK_COPY);
		put(map, "paste", KeyEvent.VK_PASTE);
		put(map, "undo", KeyEvent.VK_UNDO);
		put(map, "again", KeyEvent.VK_AGAIN);
		put(map, "find", KeyEvent.VK_FIND);
		put(map, "props", KeyEvent.VK_PROPS);
		put(map, "stop", KeyEvent.VK_STOP);
		put(map, "compose", KeyEvent.VK_COMPOSE);
		put(map, "alt_graph", KeyEvent.VK_ALT_GRAPH);
		put(map, "begin", KeyEvent.VK_BEGIN);

		return map;
	}

	private static void put(HashMap map, String key, int val) {
		map.put(key, new Integer(val));
	}

	private synchronized native void initializeLibrary()
			throws UnsatisfiedLinkError;

	private synchronized native void regHotKey(int identifier, int modifier,
			int keycode) throws UnsatisfiedLinkError;

	private synchronized native void terminate() throws UnsatisfiedLinkError;

	private synchronized native void unregHotKey(int identifier)
			throws UnsatisfiedLinkError;

	/**
	 * Checks if there's an instance with hidden window title = appName running
	 * Can be used to detect that another instance of your app is already running
	 * (so exit..)
	 * <p>
	 * @param appName = the title of the hidden window to search for
	 */
	private synchronized native boolean isRunning(String appName);
}