package petaclone.petaroux;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.BindException;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;

import jp.takayuki.logging.Logger;
import jp.takayuki.tools.HexDump;
import jp.takayuki.tools.MessageMap;
import jp.takayuki.tools.Tool;
import petaclone.InvalidDataFormatException;

/**
 * @author Takayuki Shoji
 */
public class PetaxTCPServer implements Runnable {

	/**
	 * logger
	 */
	private static final Logger logger = Logger.getLogger(PetaxTCPServer.class);

	/**
	 * fobOtO
	 */
	private static final boolean DEBUG = logger.isDebugEnabled();

	/**
	 * bZ[W\[X
	 */
	private static final MessageMap mm = MessageMap.getMessageMap();

	/** |[gԍ */
	static final int port = 20205;

	/** NȂtrue */
	private boolean isAlive;

	/** T[o\Pbg */
	private ServerSocket server;

	/** bZW[ */
	private PetaxMessenger messenger;

	/**
	 * RXgN^
	 * @param messenger bZW[
	 */
	PetaxTCPServer(PetaxMessenger messenger) {
		this.messenger = messenger;
	}

	/**
	 * T[oĂ邩Ԃ܂B
	 * @return Ăꍇtrue
	 */
	synchronized boolean getAlive() {
		return isAlive;
	}

	/**
	 * T[oɏIʒm܂B
	 */
	void quit() {
		isAlive = false;
		close();
	}

	/**
	 * @see java.lang.Runnable#run()
	 */
	public void run() {

		try {
			while (isAlive) {
				Socket socket = null;
				try {
					socket = server.accept();
				} catch (IOException e) {
					if (isAlive) {
						logger.error(//
								"IOException occurred during accept().", e); //$NON-NLS-1$
						messenger.notifyError(//
								mm.get("PetaxTCPServer.IOException.accept"), e); //$NON-NLS-1$
					}
					if (server.isBound() && !server.isClosed()) {
						continue;
					}
					return;
				}

				handleSocket(socket);
			}
			// T[o\Pbg
		} finally {
			isAlive = false;
			close();
		}
	}

	/**
	 * Mf[^܂B
	 * @param socket
	 */
	private void handleSocket(Socket socket) {

		InputStream is = null;
		OutputStream os = null;

		try {
			is = socket.getInputStream();
			byte[] buffer = readStream(is);
			if (buffer == null || buffer.length == 0) {
				return;
			}
			socket.shutdownInput();

			// Mf[^̏
			if (handleReceiveData(buffer)) {

				// ԓ
				os = socket.getOutputStream();
				sendReply(os);
			}

		} catch (IOException e) {
			logger.error("Exception occurred while handling socket.", e); //$NON-NLS-1$
			messenger.notifyError(//
					mm.get("PetaxTCPServer.IOException.handleSocket"), e); //$NON-NLS-1$

		} finally {
			try {
				if (is != null) {
					is.close();
				}
			} catch (IOException e) {
				logger.error("Exception occurred " //$NON-NLS-1$
						+ "while closing socket input stream.", e); //$NON-NLS-1$
			}
			try {
				if (os != null) {
					os.close();
				}
			} catch (IOException e) {
				logger.error("Exception occurred " //$NON-NLS-1$
						+ "while closing socket output stream.", e); //$NON-NLS-1$
			}

			try {
				socket.close();
			} catch (IOException e) {
				logger.error("Exception occurred while closing socket.", e); //$NON-NLS-1$
			}
		}
	}

	/**
	 * oCh܂B
	 * @throws BindException 
	 * @throws SocketException 
	 * @throws IOException 
	 */
	void bind() throws BindException, SocketException, IOException {
		server = new ServerSocket(port, -1, messenger.bindAddress);
		isAlive = true;
		logger.info("Bound: addr=" + messenger.bindAddress + ", port=" + port); //$NON-NLS-1$ //$NON-NLS-2$
	}

	/**
	 * T[o\Pbg܂B
	 */
	private void close() {
		try {
			if (server != null && !server.isClosed()) {
				server.close();
			}
		} catch (IOException e) {
			logger.error("Exception occurred during server.close().", e); //$NON-NLS-1$
		}
	}

	/**
	 * ҋ@ƎMB
	 * TODO [eBeB\bhɂ
	 * @param is ̓Xg[
	 * @return Mf[^
	 * @throws IOException ̓Xg[̗O
	 */
	byte[] readStream(InputStream is) throws IOException {

		// M
		byte[] buffer = null;

		byte[] int32 = new byte[4];
		if (Tool.readComplete(is, int32) < 4) {
			logger.warn("Could not read header."); //$NON-NLS-1$
			return null;
		}
		int length = Tool.byteToIntBigEndian(int32, 0);
		if (length <= 0) {
			logger.warn("Header value is invalid: " + length); //$NON-NLS-1$
			return null;
		}

		buffer = new byte[length];
		int read = Tool.readComplete(is, buffer);
		if (read < length) {
			logger.warn("Could not read data completely: " + read + "/" + length); //$NON-NLS-1$ //$NON-NLS-2$
			return null;
		}

		return buffer;
	}

	/**
	 * Mf[^܂B
	 * ʏ̃bZ[WAsݒʒmAJʒmB
	 * @param buffer Mf[^B
	 * @return łꍇtrueB
	 */
	private boolean handleReceiveData(byte[] buffer) {

		if (DEBUG)
			logger.debug(HexDump.dump(buffer));

		try {
			ByteArrayInputStream is = new ByteArrayInputStream(buffer);
			PetaxPacket packet = PetaxPacketHandler.parse(is);
			//
			if (packet.msg == null) {
				logger.debug("Could not parse received message data."); //$NON-NLS-1$
				return false;
			}

			packet.msg.protocol = messenger.getProtocol().signature;
			packet.msg.launchId = messenger.getId();

			// \
			messenger.notifyMessage(packet.msg);
			logger.debug("handleReceiveData end."); //$NON-NLS-1$
			return true;
		} catch (InvalidDataFormatException e) {
			logger.warn("Format of received data is invalid.", e); //$NON-NLS-1$
			return false;
		}
	}

	/**
	 * SENDMSGɑ΂RETURNŉ܂B
	 * @param os o̓Xg[
	 * @throws IOException
	 */
	private void sendReply(OutputStream os) throws IOException {
		// ԓpPbg쐬܂
		PetaxPacket reply = new PetaxPacket();
		reply.gmversion = PetaxPacket.GMVERSION;
		reply.msgseq = 0;
		reply.id = PetaxPacket.TYPE_RETURN;
		reply.gizmoid = ""; //$NON-NLS-1$
		reply.username = ""; //$NON-NLS-1$
		reply.hostname = ""; //$NON-NLS-1$
		reply.displayname = ""; //$NON-NLS-1$
		// TODO IPv6
		reply.ipaddress = "0.0.0.0"; //$NON-NLS-1$
		//reply.msg = null;

		// ԓ̑M
		String xml = reply.toXml();
		byte[] bytes = xml.getBytes(PetaxMessenger.ENCODING);
		byte[] int32 = new byte[4];
		Tool.writeInt32BigEndian(int32, 0, bytes.length);

		os.write(int32);
		os.write(bytes);
		os.flush();
	}
}
